# -*- coding: utf-8 -*-
u"""
Created on 2015-7-16

@author: cheng.li
"""

import unittest
import csv
import os
import copy
import pickle
import tempfile
import math
import numpy as np
from collections import deque
from PyFin.Math.Accumulators import Current
from PyFin.Math.Accumulators import Latest
from PyFin.Math.Accumulators import Shift
from PyFin.Math.Accumulators import Delta
from PyFin.Math.Accumulators import Exp
from PyFin.Math.Accumulators import MovingMax
from PyFin.Math.Accumulators import MovingArgMax
from PyFin.Math.Accumulators import MovingMin
from PyFin.Math.Accumulators import MovingArgMin
from PyFin.Math.Accumulators import MovingRank
from PyFin.Math.Accumulators import MovingQuantile
from PyFin.Math.Accumulators import MovingCount
from PyFin.Math.Accumulators import TimeMovingCount
from PyFin.Math.Accumulators import MovingCountUnique
from PyFin.Math.Accumulators import TimeMovingCountUnique
from PyFin.Math.Accumulators import MovingAllTrue
from PyFin.Math.Accumulators import MovingAnyTrue
from PyFin.Math.Accumulators import MovingAverage
from PyFin.Math.Accumulators import MovingDecay
from PyFin.Math.Accumulators import MovingPositiveAverage
from PyFin.Math.Accumulators import MovingNegativeAverage
from PyFin.Math.Accumulators import MovingPositiveDifferenceAverage
from PyFin.Math.Accumulators import MovingNegativeDifferenceAverage
from PyFin.Math.Accumulators import MovingRSI
from PyFin.Math.Accumulators import MovingSum
from PyFin.Math.Accumulators import MovingCountedPositive
from PyFin.Math.Accumulators import MovingCountedNegative
from PyFin.Math.Accumulators import MovingVariance
from PyFin.Math.Accumulators import MovingStandardDeviation
from PyFin.Math.Accumulators import MovingNegativeVariance
from PyFin.Math.Accumulators import MovingCorrelation
from PyFin.Math.Accumulators import MovingProduct
from PyFin.Math.Accumulators import MACD
from PyFin.Math.Accumulators import XAverage
from PyFin.Math.Accumulators import MovingResidue


class TestStatefulAccumulators(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        self.sample = np.random.randn(10000)

    def testParse(self):
        exp = TimeMovingCount(10, "x")
        self.assertEqual(exp.window, 10)

        exp = TimeMovingCount(12., "x")
        self.assertEqual(exp.window, 12)

        exp = TimeMovingCount("2D", "x")
        self.assertEqual(exp.window, 2. * 24 * 3600)

        exp = TimeMovingCount("3H", "x")
        self.assertEqual(exp.window, 3. * 3600)

        exp = TimeMovingCount("4T", "x")
        self.assertEqual(exp.window, 4. * 60)

        exp = TimeMovingCount("5S", "x")
        self.assertEqual(exp.window, 5.)

    def testCurrentValueHolderWithOtherValueHolder(self):
        test1 = Exp(Current('x'))
        test1.push({'x': 2.0})
        self.assertAlmostEqual(test1.value, math.exp(2.0))

        test1.push({'x': np.nan})
        self.assertTrue(math.isnan(test1.value))

    def testLatestValueHolderWithOtherValueHolder(self):
        test1 = Exp(Latest('x'))
        test1.push({'x': 2.0})
        self.assertAlmostEqual(test1.value, math.exp(2.0))

        test1.push({'x': np.nan})
        self.assertAlmostEqual(test1.value, math.exp(2.0))

    def testLatestValueHolderDeepCopy(self):

        latest = Latest('x')
        copied = copy.deepcopy(latest)

        self.assertTrue(math.isnan(copied.value))

        latest.push({'x': 1.})
        copied = copy.deepcopy(latest)
        self.assertTrue(copied.value, 1.)

    def testLatestValueHolderPickle(self):

        latest = Latest('x')

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(latest, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertTrue(math.isnan(pickled.result()))

        os.unlink(f.name)

        latest.push({'x': 1.})

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(latest, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertTrue(pickled.value, 1.)

        os.unlink(f.name)

    def testShiftValueHolder(self):
        ma = MovingAverage(10, 'close')

        with self.assertRaises(ValueError):
            _ = Shift(0, ma)

        test = Shift(1, ma)

        test.push(dict(close=2.0))
        ma.push(dict(close=2.0))
        previous = ma.result()
        test.push(dict(close=5.0))
        ma.push(dict(close=5.0))

        self.assertAlmostEqual(previous, test.result())

        previous = ma.result()
        test.push(dict(close=10.0))
        self.assertAlmostEqual(previous, test.result())

    def testDeltaValueHolder(self):
        ma = MovingAverage(10, 'close')

        with self.assertRaises(ValueError):
            _ = Delta(window=0, x=ma)

        test = Delta(2, ma)

        test.push(dict(close=2.0))
        ma.push(dict(close=2.0))
        p_previous = ma.result()
        test.push(dict(close=5.0))
        ma.push(dict(close=5.0))
        previous = ma.result()
        test.push(dict(close=7.0))
        ma.push(dict(close=7.0))

        self.assertAlmostEqual(ma.result() - p_previous, test.result())

        ma.push(dict(close=10.0))
        test.push(dict(close=10.0))
        self.assertAlmostEqual(ma.result() - previous, test.result())

    def testShiftValueHolderDeepcopy(self):
        ma = Latest('close')
        test = Shift(2, ma)

        test.push(dict(close=2.0))
        test.push(dict(close=3.0))
        test.push(dict(close=4.0))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testShiftValueHolderPickle(self):
        ma = Latest('close')
        test = Shift(2, ma)

        test.push(dict(close=2.0))
        test.push(dict(close=3.0))
        test.push(dict(close=4.0))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(test, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingAverager(self):
        window = 120
        total = 2500

        mv = MovingAverage(window, 'z')
        runningSum = 0.0
        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            runningSum += value
            if i >= window:
                runningSum -= con[0]
                con = con[1:]

            if i >= window - 1:
                expected = runningSum / window
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Average expected:   {1:f}\n"
                                                                 "Average calculated: {2:f}".format(i, expected,
                                                                                                    calculated))

    def testMovingDecay(self):
        window = 80
        total = 2500

        mv = MovingDecay(window, 'z')

        def calculate_decay(con, k):
            s = k - len(con) + 1
            c = (k + s) * (k - s + 1) / 2.
            sum_value = 0.
            for w in range(s, k + 1):
                i = w - s
                sum_value += w * con[i]
            return sum_value / c

        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            if i >= window:
                con = con[1:]

            if i >= 1:
                expected = calculate_decay(con, window)
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Average expected:   {1:f}\n"
                                                                 "Average calculated: {2:f}".format(i, expected,
                                                                                                    calculated))

    def testMovingAverageDeepcopy(self):
        window = 20
        mv = MovingAverage(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

        for i in range(3, 40):
            mv.push(dict(x=float(i)))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

    def testMovingAveragePickle(self):
        window = 20
        mv = MovingAverage(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(mv, f)
        f.close()
        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

        for i in range(3, 40):
            mv.push(dict(x=float(i)))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

    def testMovingPositiveAverager(self):
        dirName = os.path.dirname(os.path.abspath(__file__))
        filePath = os.path.join(dirName, 'data/averagepostiveandnegative_random.csv')

        window = 20
        mv = MovingPositiveAverage(window, 'z')
        self.assertAlmostEqual(mv.result(), 0.0, 15)

        with open(filePath, 'r') as fileHandler:
            reader = csv.reader(fileHandler)

            for i, row in enumerate(reader):
                if i == 0:
                    continue
                mv.push(dict(z=float(row[1])))
                if i >= window:
                    expected = float(row[6])
                    calculated = mv.result()
                    self.assertAlmostEqual(calculated, expected, 8, "at index {0:d}\n"
                                                                    "Positive average expected:   {1:f}\n"
                                                                    "positive average calculated: {2:f}".format(i,
                                                                                                                expected,
                                                                                                                calculated))

    def testMovingPositiveAverageDeepcopy(self):
        mv = MovingPositiveAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(mv.value, copied.value)

        data = np.random.randn(20)

        for d in data:
            mv.push(dict(x=d))
            copied.push(dict(x=d))

            self.assertAlmostEqual(mv.value, copied.value)

    def testMovingPositiveAveragePickle(self):
        mv = MovingPositiveAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        self.assertAlmostEqual(mv.value, pickled.value)

        data = np.random.randn(20)

        for d in data:
            mv.push(dict(x=d))
            pickled.push(dict(x=d))

            self.assertAlmostEqual(mv.value, pickled.value)

    def testMovingPositiveDifferenceAverage(self):
        window = 10
        mv = MovingPositiveDifferenceAverage(window, 'x')

        last = self.sample[0]
        diff_list = []
        for i, value in enumerate(self.sample):
            mv.push(dict(x=value))
            diff = value - last
            if i > 0:
                diff_list.append(diff)
            last = value
            pos_list = np.maximum(diff_list[-window:], 0.)

            if i > 1:
                expected = 0. if np.isnan(np.mean(pos_list)) else np.mean(pos_list)
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 8, "at index {0:d}\n"
                                                                "Positive average expected:   {1:f}\n"
                                                                "positive average calculated: {2:f}".format(i,
                                                                                                            expected,
                                                                                                            calculated))

    def testMovingPositiveDifferenceAverageDeepcopy(self):
        mv = MovingPositiveDifferenceAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(mv.value, copied.value)

        data = np.random.randn(20)
        for d in data:
            mv.push(dict(x=d))
            copied.push(dict(x=d))

            self.assertAlmostEqual(mv.value, copied.value)

    def testMovingPositiveDifferenceAveragePickle(self):
        mv = MovingPositiveDifferenceAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        self.assertAlmostEqual(mv.value, pickled.value)

        data = np.random.randn(20)

        for d in data:
            mv.push(dict(x=d))
            pickled.push(dict(x=d))

            self.assertAlmostEqual(mv.value, pickled.value)

    def testMovingNegativeDifferenceAverage(self):
        window = 10
        mv = MovingNegativeDifferenceAverage(window, 'x')

        last = self.sample[0]
        diff_list = []
        for i, value in enumerate(self.sample):
            mv.push(dict(x=value))
            diff = value - last
            if i > 0:
                diff_list.append(diff)
            last = value
            neg_list = np.minimum(diff_list[-window:], 0.)

            if i > 0 and len(neg_list) > 0:
                expected = 0. if np.isnan(np.mean(neg_list)) else np.mean(neg_list)
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 8, "at index {0:d}\n"
                                                                "negative average expected:   {1:f}\n"
                                                                "negative average calculated: {2:f}".format(i,
                                                                                                            expected,
                                                                                                            calculated))

    def testMovingNegativeDifferenceAverageDeepcopy(self):
        mv = MovingNegativeDifferenceAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(mv.value, copied.value)

        data = np.random.randn(20)
        for d in data:
            mv.push(dict(x=d))
            copied.push(dict(x=d))

            self.assertAlmostEqual(mv.value, copied.value)

    def testMovingNegativeDifferenceAveragePickle(self):
        mv = MovingNegativeDifferenceAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        self.assertAlmostEqual(mv.value, pickled.value)

        data = np.random.randn(20)

        for d in data:
            mv.push(dict(x=d))
            pickled.push(dict(x=d))

            self.assertAlmostEqual(mv.value, pickled.value)

    def testMovingRSI(self):
        window = 10
        rsi = MovingRSI(window, 'x')
        pos_avg = MovingPositiveDifferenceAverage(window, 'x')
        neg_avg = MovingNegativeDifferenceAverage(window, 'x')
        for i, value in enumerate(self.sample):
            rsi.push(dict(x=value))
            pos_avg.push(dict(x=value))
            neg_avg.push(dict(x=value))

            calculated = rsi.value
            expected = pos_avg.value / (pos_avg.value - neg_avg.value) * 100
            if i > 0:
                self.assertAlmostEqual(calculated, expected, 8, "at index {0:d}\n"
                                                                "negative average expected:   {1:f}\n"
                                                                "negative average calculated: {2:f}".format(i,
                                                                                                            expected,
                                                                                                            calculated))

    def testMovingRSIDeepcopy(self):
        mv = MovingRSI(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(mv.value, copied.value)

    def testMovingRSIPickle(self):
        mv = MovingRSI(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        self.assertAlmostEqual(mv.value, pickled.value)

    def testMovingNegativeAverager(self):
        dirName = os.path.dirname(os.path.abspath(__file__))
        filePath = os.path.join(dirName, 'data/averagepostiveandnegative_random.csv')

        window = 20
        mv = MovingNegativeAverage(window, 'z')
        self.assertAlmostEqual(mv.result(), 0.0, 15)

        with open(filePath, 'r') as fileHandler:
            reader = csv.reader(fileHandler)

            for i, row in enumerate(reader):
                if i == 0:
                    continue
                mv.push(dict(z=float(row[1])))
                if i >= window:
                    expected = float(row[7])
                    calculated = mv.result()
                    self.assertAlmostEqual(calculated, expected, 8, "at index {0:d}\n"
                                                                    "Positive average expected:   {1:f}\n"
                                                                    "positive average calculated: {2:f}".format(i,
                                                                                                                expected,
                                                                                                                calculated))

    def testMovingNegativeAverageDeepcopy(self):
        mv = MovingNegativeAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(mv.value, copied.value)

    def testMovingNegativeAveragePickle(self):
        mv = MovingNegativeAverage(3, 'x')
        mv.push(dict(x=1.))
        mv.push(dict(x=-1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        self.assertAlmostEqual(mv.value, pickled.value)

    def testMovingSum(self):
        window = 120
        total = 2500

        mv = MovingSum(window, 'z')
        runningSum = 0.0
        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            runningSum += value

            if i >= window:
                runningSum -= con[0]
                con = con[1:]

            if i >= window - 1:
                expected = runningSum
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Sum expected:   {1:f}\n"
                                                                 "Sum calculated: {2:f}".format(i, expected,
                                                                                                calculated))

    def testMovingSumDeepcopy(self):
        ms = MovingSum(3, 'x')

        ms.push(dict(x=1.))
        ms.push(dict(x=2.))
        ms.push(dict(x=3.))

        copied = copy.deepcopy(ms)
        self.assertAlmostEqual(copied.value, ms.value)

    def testMovingSumPickle(self):
        ms = MovingSum(3, 'x')

        ms.push(dict(x=1.))
        ms.push(dict(x=2.))
        ms.push(dict(x=3.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(ms, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(ms.value, pickled.value)
        os.unlink(f.name)

    def testMovingMax(self):
        window = 120

        mv = MovingMax(window, 'z')
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mv.push(dict(z=value))
            runningMax = max(con)

            expected = runningMax
            calculated = mv.result()
            self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                             "Max expected:   {1:f}\n"
                                                             "Max calculated: {2:f}".format(i, expected, calculated))

    def testMovingArgMax(self):
        window = 120

        mv = MovingArgMax(window, 'z')
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mv.push(dict(z=value))
            runningMax = max(con)

            expected = len(con) - np.where(np.array(con) == runningMax)[0][0] - 1.
            calculated = mv.result()
            self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                             "Max expected at:   {1:f}\n"
                                                             "Max calculated at: {2:f}".format(i, expected, calculated))

    def testMovingArgMin(self):
        window = 120

        mv = MovingArgMin(window, 'z')
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mv.push(dict(z=value))
            runningMin = min(con)

            expected = len(con) - np.where(np.array(con) == runningMin)[0][0] - 1.
            calculated = mv.result()
            self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                             "Min expected at:   {1:f}\n"
                                                             "Min calculated at: {2:f}".format(i, expected, calculated))

    def testMovingMaxDeepcopy(self):
        test = MovingMax(3, 'close')

        test.push(dict(close=3.0))
        test.push(dict(close=4.0))
        test.push(dict(close=2.0))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingMaxPickle(self):
        test = MovingMax(3, 'close')

        test.push(dict(close=3.0))
        test.push(dict(close=4.0))
        test.push(dict(close=2.0))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(test, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingMinimum(self):
        window = 120

        mv = MovingMin(window, 'z')
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mv.push(dict(z=value))
            runningMin = min(con)

            expected = runningMin
            calculated = mv.result()
            self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                             "Min expected:   {1:f}\n"
                                                             "Min calculated: {2:f}".format(i, expected, calculated))

    def testMovingMinimumDeepcopy(self):
        test = MovingMin(3, 'close')

        test.push(dict(close=3.0))
        test.push(dict(close=4.0))
        test.push(dict(close=2.0))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingMinimumPickle(self):
        test = MovingMin(3, 'close')

        test.push(dict(close=3.0))
        test.push(dict(close=4.0))
        test.push(dict(close=2.0))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(test, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingRank(self):
        window = 10

        mq = MovingRank(window, 'z')
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mq.push(dict(z=value))

            if i >= 1:
                calculated = mq.result()
                sorted_con = sorted(con)
                expected = sorted_con.index(value)
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Rank expected:   {1:f}\n"
                                                                 "Rank calculated: {2:f}".format(i,
                                                                                                 expected,
                                                                                                 calculated))

    def testMovingQuantile(self):
        window = 10

        mq = MovingQuantile(window, 'z')
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mq.push(dict(z=value))

            if i >= 1:
                calculated = mq.result()
                sorted_con = sorted(con)
                expected = sorted_con.index(value) / (len(sorted_con) - 1.)
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Quantile expected:   {1:f}\n"
                                                                 "Quantile calculated: {2:f}".format(i, expected,
                                                                                                     calculated))

    def testMovingQuantileDeepcopy(self):
        test = MovingQuantile(10, 'close')

        test.push(dict(close=3.0))
        test.push(dict(close=4.0))
        test.push(dict(close=2.0))
        test.push(dict(close=3.5))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingQuantilePickle(self):
        test = MovingQuantile(10, 'close')

        test.push(dict(close=3.0))
        test.push(dict(close=4.0))
        test.push(dict(close=2.0))
        test.push(dict(close=3.5))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(test, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingCount(self):
        window = 10
        mat = MovingCount(window, Latest('x'))
        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mat.push(dict(x=value))

            calculated = mat.result()
            expected = len(con)
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountRight(self):
        window = 60
        mat = TimeMovingCount(window, Latest("x"))
        values = np.random.randn(2500)
        stamps = np.cumsum(np.random.randint(0, 10, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) < window
            expected = np.sum(time_diff[:i+1])
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountBoth(self):
        window = 60
        mat = TimeMovingCount(window, Latest("x"), "both")
        values = np.random.randn(2500)
        stamps = np.cumsum(np.random.randint(0, 10, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) <= window
            expected = np.sum(time_diff[:i+1])
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountLeft(self):
        window = 60
        mat = TimeMovingCount(window, Latest("x"), "left")
        values = np.random.randn(2500)
        stamps = np.cumsum(np.random.randint(0, 10, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) <= window
            expected = np.sum(time_diff[:i])
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountNeither(self):
        window = 60
        mat = TimeMovingCount(window, Latest("x"), "neither")
        values = np.random.randn(2500)
        stamps = np.cumsum(np.random.randint(0, 10, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) < window
            expected = np.sum(time_diff[:i])
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testMovingCountUnique(self):
        window = 10
        mat = MovingCountUnique(window, Latest('x'))
        total = np.random.randint(0, 10, 2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mat.push(dict(x=value))

            calculated = mat.result()
            expected = len(np.unique(con))
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountUniqueRight(self):
        window = 60
        mat = TimeMovingCountUnique(window, Latest("x"), "right")
        values = np.random.randint(0, 10, 2500)
        stamps = np.cumsum(np.random.randint(1, 10, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) < window
            time_diff[i+1:] = False
            expected = len(np.unique(values[time_diff]))
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountUniqueBoth(self):
        window = 60
        mat = TimeMovingCountUnique(window, Latest("x"), "both")
        values = np.random.randint(0, 10, 2500)
        stamps = np.cumsum(np.random.randint(1, 5, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) <= window
            time_diff[i+1:] = False
            expected = len(np.unique(values[time_diff]))
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountUniqueLeft(self):
        window = 60
        mat = TimeMovingCountUnique(window, Latest("x"), "left")
        values = np.random.randint(0, 10, 2500)
        stamps = np.cumsum(np.random.randint(1, 5, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) <= window
            time_diff[i:] = False
            expected = len(np.unique(values[time_diff]))
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testTimeMovingCountUniqueNeither(self):
        window = 60
        mat = TimeMovingCountUnique(window, Latest("x"), "neither")
        values = np.random.randint(0, 10, 2500)
        stamps = np.cumsum(np.random.randint(1, 5, 2500))
        for i, (value, stamp) in enumerate(zip(values, stamps)):
            mat.push(dict(x=value, stamp=stamp))

            calculated = mat.result()
            time_diff = (stamp - stamps) < window
            time_diff[i:] = False
            expected = len(np.unique(values[time_diff]))
            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "expected:   {1}\n"
                                                   "calculated: {2}".format(i, expected, calculated))

    def testMovingAllTrue(self):
        window = 3
        underlying = Latest('x') > 0.

        mat = MovingAllTrue(window, underlying)

        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mat.push(dict(x=value))

            calculated = mat.result()
            expected = np.all(np.array(con) > 0.)

            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "Quantile expected:   {1}\n"
                                                   "Quantile calculated: {2}".format(i, expected, calculated))

    def testMovingAllTrueDeepcopy(self):
        test = MovingAllTrue(3, Latest('x') > 0.)

        test.push(dict(x=-3.0))
        test.push(dict(x=4.0))
        test.push(dict(x=2.0))
        test.push(dict(x=3.5))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingAllTruePickle(self):
        test = MovingAllTrue(3, Latest('x') > 0.)

        test.push(dict(x=3.0))
        test.push(dict(x=4.0))
        test.push(dict(x=2.0))
        test.push(dict(x=-3.5))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(test, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingAnyTrue(self):
        window = 3

        underlying = Latest('x') > 0.

        mat = MovingAnyTrue(window, underlying)

        total = np.random.randn(2500)
        con = deque(maxlen=window)
        for i, value in enumerate(total):
            value = float(value)
            con.append(value)
            mat.push(dict(x=value))

            calculated = mat.result()
            expected = np.any(np.array(con) > 0.)

            self.assertEqual(calculated, expected, "at index {0}\n"
                                                   "Quantile expected:   {1}\n"
                                                   "Quantile calculated: {2}".format(i, expected, calculated))

    def testMovingAnyTrueDeepcopy(self):
        test = MovingAnyTrue(3, Latest('x') > 0.)

        test.push(dict(x=-3.0))
        test.push(dict(x=-4.0))
        test.push(dict(x=2.0))
        test.push(dict(x=-3.5))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingAnyTruePickle(self):
        test = MovingAnyTrue(3, Latest('x') > 0.)

        test.push(dict(x=3.0))
        test.push(dict(x=-4.0))
        test.push(dict(x=-2.0))
        test.push(dict(x=-3.5))

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(test, f)
        f.close()
        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingCountedPositive(self):
        window = 120
        total = 2500

        values = [1.0 if i % 2 else -1.0 for i in range(total)]
        mv = MovingCountedPositive(window, 'z')
        runningCount = 0
        con = []
        for i, value in enumerate(values):
            if i % 2:
                runningCount += 1
            con.append(i)
            mv.push(dict(z=value))

            if i >= window:
                if con[0] % 2:
                    runningCount -= 1
                con = con[1:]

            if i >= window - 1:
                expected = runningCount
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Counted positive expected:   {1:f}\n"
                                                                 "Counted positive calculated: {2:f}".format(i,
                                                                                                             expected,
                                                                                                             calculated))

    def testMovingCountedPositiveDeepcopy(self):
        test = MovingCountedPositive(3, 'x')

        test.push(dict(x=-3.0))
        test.push(dict(x=-4.0))
        test.push(dict(x=2.0))
        test.push(dict(x=-3.5))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingCountedPositivePickle(self):
        test = MovingCountedPositive(3, 'x')

        test.push(dict(x=3.0))
        test.push(dict(x=-4.0))
        test.push(dict(x=-2.0))
        test.push(dict(x=-3.5))

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(test, f)
        f.close()
        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingCountedNegative(self):
        window = 120
        total = 2500

        values = [1.0 if i % 2 else -1.0 for i in range(total)]
        mv = MovingCountedNegative(window, 'z')
        runningCount = 0
        con = []
        for i, value in enumerate(values):
            if not i % 2:
                runningCount += 1
            con.append(i)
            mv.push(dict(z=value))

            if i >= window:
                if not con[0] % 2:
                    runningCount -= 1
                con = con[1:]

            if i >= window - 1:
                expected = runningCount
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Counted negative expected:   {1:f}\n"
                                                                 "Counted negative calculated: {2:f}".format(i,
                                                                                                             expected,
                                                                                                             calculated))

    def testMovingCountedNegativeDeepcopy(self):
        test = MovingCountedNegative(3, 'x')

        test.push(dict(x=-3.0))
        test.push(dict(x=-4.0))
        test.push(dict(x=2.0))
        test.push(dict(x=-3.5))

        copied = copy.deepcopy(test)
        self.assertAlmostEqual(copied.value, test.value)

    def testMovingCountedNegativePickle(self):
        test = MovingCountedNegative(3, 'x')

        test.push(dict(x=3.0))
        test.push(dict(x=-4.0))
        test.push(dict(x=-2.0))
        test.push(dict(x=-3.5))

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(test, f)
        f.close()
        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(test.value, pickled.value)
        os.unlink(f.name)

    def testMovingVariancer(self):
        window = 120
        total = 2500

        # Test moving population variance
        mv = MovingVariance(window, 'z', isPopulation=True)
        runningSum = 0.0
        runningSumSquare = 0.0
        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            runningSum += value
            runningSumSquare += value * value

            if i >= window:
                runningSum -= con[0]
                runningSumSquare -= con[0] * con[0]
                con = con[1:]

            if i >= window - 1:
                expected = (runningSumSquare - runningSum * runningSum / window) / window
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Var expected:   {1:f}\n"
                                                                 "Var calculated: {2:f}".format(i, expected,
                                                                                                calculated))

        # Test moving sample variance
        mv = MovingVariance(window, 'z', isPopulation=False)
        runningSum = 0.0
        runningSumSquare = 0.0
        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            runningSum += value
            runningSumSquare += value * value

            if i == 0:
                self.assertTrue(np.isnan(mv.result()))

            if i >= window:
                runningSum -= con[0]
                runningSumSquare -= con[0] * con[0]
                con = con[1:]

            length = window if i >= window else (i + 1)
            if i >= window - 1:
                expected = (runningSumSquare - runningSum * runningSum / length) / (length - 1)
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Var expected:   {1:f}\n"
                                                                 "Var calculated: {2:f}".format(i, expected,
                                                                                                calculated))

    def testMovingVarianceDeepcopy(self):
        window = 20
        mv = MovingVariance(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

        for i in range(3, 40):
            mv.push(dict(x=float(i)))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

    def testMovingVariancePickle(self):
        window = 20
        mv = MovingVariance(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

        for i in range(3, 40):
            mv.push(dict(x=float(i)))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

    def testMovingStandardDeviation(self):
        window = 120
        total = 2500

        # Test moving population variance
        mv = MovingStandardDeviation(window, 'z', isPopulation=True)
        runningSum = 0.0
        runningSumSquare = 0.0
        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            runningSum += value
            runningSumSquare += value * value

            if i >= window:
                runningSum -= con[0]
                runningSumSquare -= con[0] * con[0]
                con = con[1:]

            if i >= window - 1:
                expected = math.sqrt((runningSumSquare - runningSum * runningSum / window) / window)
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Var expected:   {1:f}\n"
                                                                 "Var calculated: {2:f}".format(i, expected,
                                                                                                calculated))

        # Test moving sample variance
        mv = MovingVariance(window, 'z', isPopulation=False)
        runningSum = 0.0
        runningSumSquare = 0.0
        con = []
        for i in range(total):
            value = float(i)
            con.append(value)
            mv.push(dict(z=value))
            runningSum += value
            runningSumSquare += value * value

            if i == 0:
                self.assertTrue(np.isnan(mv.result()))

            if i >= window:
                runningSum -= con[0]
                runningSumSquare -= con[0] * con[0]
                con = con[1:]

            length = window if i >= window else (i + 1)
            if i >= window - 1:
                expected = (runningSumSquare - runningSum * runningSum / length) / (length - 1)
                calculated = mv.result()
                self.assertAlmostEqual(calculated, expected, 15, "at index {0:d}\n"
                                                                 "Var expected:   {1:f}\n"
                                                                 "Var calculated: {2:f}".format(i, expected,
                                                                                                calculated))

    def testMovingStandardDeviationDeepcopy(self):
        window = 20
        mv = MovingStandardDeviation(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

        for i in range(3, 40):
            mv.push(dict(x=float(i)))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

    def testMovingStandardDeviationPickle(self):
        window = 20
        mv = MovingStandardDeviation(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

        for i in range(3, 40):
            mv.push(dict(x=float(i)))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

    def testMovingNegativeVariancer(self):
        # test without enough negative value
        mv = MovingNegativeVariance(20, 'z', isPopulation=True)
        mv.push(dict(z=20.))

        self.assertTrue(np.isnan(mv.result()))

        mv = MovingNegativeVariance(20, 'z', isPopulation=False)
        mv.push(dict(z=20.))
        mv.push(dict(z=-20.))

        self.assertTrue(np.isnan(mv.result()))

        dirName = os.path.dirname(os.path.abspath(__file__))
        filePath = os.path.join(dirName, 'data/negativevariance.csv')

        window = 20
        mv = MovingNegativeVariance(window, 'z')

        with open(filePath, 'r') as fileHandler:
            reader = csv.reader(fileHandler)
            for i, row in enumerate(reader):
                if i == 0:
                    continue
                mv.push(dict(z=float(row[1])))

                if mv._runningNegativeCount == 1:
                    self.assertTrue(np.isnan(mv.result()))

                if i >= window:
                    expected = float(row[6])
                    calculated = mv.result()
                    self.assertAlmostEqual(calculated, expected, 7, "at index {0:d}\n"
                                                                    "Negative variance expected:   {1:f}\n"
                                                                    "Negative variance calculated: {2:f}".format(i,
                                                                                                                 expected,
                                                                                                                 calculated))

    def testMovingNegativeVarianceDeepcopy(self):
        window = 20
        mv = MovingNegativeVariance(window, 'x')

        mv.push(dict(x=-1.))
        mv.push(dict(x=-2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

        datas = np.random.randn(40)
        for i in datas:
            mv.push(dict(x=i))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

    def testMovingNegativeVariancePickle(self):
        window = 20
        mv = MovingNegativeVariance(window, 'x')

        mv.push(dict(x=-1.))
        mv.push(dict(x=-2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

        datas = np.random.randn(40)
        for i in datas:
            mv.push(dict(x=i))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

    def testMovingCorrelation(self):
        dirName = os.path.dirname(os.path.abspath(__file__))
        filePath = os.path.join(dirName, 'data/correlation.csv')

        window = 120
        mv = MovingCorrelation(window, 'z', 't')

        with open(filePath, 'r') as fileHandler:
            reader = csv.reader(fileHandler)

            for i, row in enumerate(reader):
                if i == 0:
                    continue
                mv.push(dict(z=float(row[0]), t=float(row[1])))

                if i == 1:
                    self.assertTrue(np.isnan(mv.result()))

                if i >= window:
                    expected = float(row[2])
                    calculated = mv.result()
                    self.assertAlmostEqual(calculated, expected, 8, "at index {0:d}\n"
                                                                    "Correlation expected:   {1:f}\n"
                                                                    "Correlation calculated: {2:f}".format(i, expected,
                                                                                                           calculated))

    def testMovingProduct(self):
        window = 10
        mp = MovingProduct(window, 'x')

        con = []
        for i, value in enumerate(self.sample):
            con.append(value)
            mp.push(dict(x=value))
            if i >= window:
                con = con[1:]
            if i >= window - 1:
                expected = np.product(con)
                calculated = mp.result()
                self.assertAlmostEqual(expected, calculated, 8, "at index {0:d}\n"
                                                                "product expected:   {1:f}\n"
                                                                "product calculated: {2:f}".format(i, expected,
                                                                                                   calculated))

    def testMovingProductDeepcopy(self):
        window = 10
        mv = MovingProduct(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

        for i in range(3, 20):
            mv.push(dict(x=float(i)))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

    def testMovingProductPickle(self):
        window = 10
        mv = MovingProduct(window, 'x')

        mv.push(dict(x=1.))
        mv.push(dict(x=2.))

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(mv, f)
        f.close()
        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

        for i in range(3, 20):
            mv.push(dict(x=float(i)))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

    def testMACD(self):
        macd = MACD(5, 10, 'close')
        short_average = XAverage(5, 'close')
        long_average = XAverage(10, 'close')

        for i, value in enumerate(self.sample):
            macd.push(dict(close=value))
            short_average.push(dict(close=value))
            long_average.push(dict(close=value))
            expected = short_average.result() - long_average.result()

            calculated = macd.result()
            self.assertAlmostEqual(expected, calculated, 10, "at index {0:d}\n"
                                                             "expected x-average:   {1:f}\n"
                                                             "calculated x-average: {2:f}".format(i,
                                                                                                  expected,
                                                                                                  calculated))

    def tesMACDDeepcopy(self):
        mv = MACD(5, 10, 'x')

        mv.push(dict(x=-1.))
        mv.push(dict(x=-2.))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

        datas = np.random.randn(40)
        for i in datas:
            mv.push(dict(x=i))

        copied = copy.deepcopy(mv)
        self.assertAlmostEqual(copied.value, mv.value)

    def testMACDPickle(self):
        mv = MACD(5, 10, 'x')

        mv.push(dict(x=-1.))
        mv.push(dict(x=-2.))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

        datas = np.random.randn(40)
        for i in datas:
            mv.push(dict(x=i))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mv, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            self.assertAlmostEqual(mv.value, pickled.value)
        os.unlink(f.name)

    def testEMAMACD(self):
        fast = 5
        slow = 10
        ema_window = 10
        macd = MACD(fast, slow, 'close')
        ema_macd = XAverage(ema_window, macd)

        macd_diff = macd - ema_macd

        for i, value in enumerate(self.sample):
            macd.push(dict(close=value))
            ema_macd.push(dict(close=value))
            macd_diff.push(dict(close=value))
            expected = macd.value - ema_macd.value
            calculated = macd_diff.value
            self.assertAlmostEqual(expected, calculated, 10, "at index {0:d}\n"
                                                             "expected ema macd diff:   {1:f}\n"
                                                             "calculated ema macd diff: {2:f}".format(i, expected,
                                                                                                      calculated))

    def testMovingResidue(self):
        window = 100
        x_data = np.random.randn(10000)

        mr = MovingResidue(window, x='x', y='y')

        for i, value in enumerate(zip(x_data, self.sample)):
            x = value[0]
            y = value[1]

            mr.push(dict(x=x, y=y))

            if i >= window - 1:
                series_x = x_data[i - window + 1:i + 1]
                series_y = self.sample[i - window + 1:i + 1]

                calculated_res = mr.result()
                expected_res = y - np.dot(series_x, series_y) / np.dot(series_x, series_x) * x
                self.assertAlmostEqual(expected_res, calculated_res, 8, "at index of {0:d}\n"
                                                                        "expected res:  {1:f}\n"
                                                                        "calculated res:{2:f}".format(i,
                                                                                                      expected_res,
                                                                                                      calculated_res))

    def testMovingResidueDeepcopy(self):
        window = 10
        mr = MovingResidue(window, x='x', y='y')

        mr.push(dict(x=-1., y=0))
        mr.push(dict(x=-2., y=-1))
        mr.push(dict(x=-3., y=-2))
        mr.push(dict(x=-4., y=-3))

        copied = copy.deepcopy(mr)

        datas = np.random.randn(40)
        for i in datas:
            mr.push(dict(x=i, y=i + 1.))
            copied.push(dict(x=i, y=i + 1))

        self.assertAlmostEqual(copied.value, mr.value)

    def testMovingResiduePickle(self):
        window = 10
        mr = MovingResidue(window, x='x', y='y')

        mr.push(dict(x=-1., y=0))
        mr.push(dict(x=-2., y=-1))
        mr.push(dict(x=-3., y=-2))

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(mr, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        datas = np.random.randn(40)
        for i in datas:
            mr.push(dict(x=i, y=i + 1.))
            pickled.push(dict(x=i, y=i + 1.))

        self.assertAlmostEqual(mr.value, pickled.value)
        os.unlink(f.name)

    def testShiftStr(self):
        s = Shift(2, Latest('roe'))
        self.assertEqual("\mathrm{Shift}(''\\text{roe}'', 2)", str(s))

    def testMovingMaxStr(self):
        s = MovingMax(10, 'roe')
        self.assertEqual("\mathrm{MMax}(10, ''\\text{roe}'')", str(s))

        s = MovingMax(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MMax}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testMovingMinStr(self):
        s = MovingMin(10, 'roe')
        self.assertEqual("\mathrm{MMin}(10, ''\\text{roe}'')", str(s))

        s = MovingMin(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MMin}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testMovingQuantileStr(self):
        s = MovingQuantile(10, 'roe')
        self.assertEqual("\mathrm{MQuantile}(10, ''\\text{roe}'')", str(s))

        s = MovingQuantile(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MQuantile}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testMovingCountStr(self):
        s = MovingCount(10, 'roe')
        self.assertEqual("\mathrm{MCount}(10, ''\\text{roe}'')", str(s))

        s = MovingCount(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MCount}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testTimeMovingCountStr(self):
        s = TimeMovingCount(10, 'roe')
        self.assertEqual("\mathrm{TimeMCount}(10, ''\\text{roe}'', b'right')", str(s))

        s = TimeMovingCount(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{TimeMCount}(10, ''\\text{roe}'' + ''\\text{y}'', b'right')", str(s))

    def testMovingCountUniqueStr(self):
        s = MovingCountUnique(10, 'roe')
        self.assertEqual("\mathrm{MCountUnique}(10, ''\\text{roe}'')", str(s))

        s = MovingCountUnique(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MCountUnique}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testTimeMovingCountUniqueStr(self):
        s = TimeMovingCountUnique(10, 'roe')
        self.assertEqual("\mathrm{TimeMCountUnique}(10, ''\\text{roe}'', b'right')", str(s))

        s = TimeMovingCountUnique(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{TimeMCountUnique}(10, ''\\text{roe}'' + ''\\text{y}'', b'right')", str(s))

    def testMovingAllTrueStr(self):
        s = MovingAllTrue(10, 'roe')
        self.assertEqual("\mathrm{MAllTrue}(10, ''\\text{roe}'')", str(s))

        s = MovingAllTrue(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MAllTrue}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testMovingAnyTrueStr(self):
        s = MovingAnyTrue(10, 'roe')
        self.assertEqual("\mathrm{MAnyTrue}(10, ''\\text{roe}'')", str(s))

        s = MovingAnyTrue(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MAnyTrue}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testMovingSumStr(self):
        s = MovingSum(10, 'roe')
        self.assertEqual("\mathrm{MSum}(10, ''\\text{roe}'')", str(s))

        s = MovingSum(10, Latest('roe') + Latest('y'))
        self.assertEqual("\mathrm{MSum}(10, ''\\text{roe}'' + ''\\text{y}'')", str(s))

    def testMovingResidueResStr(self):
        s = MovingResidue(10, x='roe', y='y')
        self.assertEqual("\mathrm{Res}(10, ''\\text{roe}'', ''\\text{y}'')", str(s))

        s = MovingResidue(10, x=Latest('roe'), y=Latest('y'))
        self.assertEqual("\mathrm{Res}(10, ''\\text{roe}'', ''\\text{y}'')", str(s))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-30

@author: cheng.li
"""

from PyFin.Enums._TimeUnits cimport TimeUnits as tu

cpdef enum TimeUnits:
    BDays = tu.BDays
    Days = tu.Days
    Weeks = tu.Weeks
    Months = tu.Months
    Years = tu.Years


# -*- coding: utf-8 -*-
u"""
Created on 2017-6-7

@author: cheng.li
"""

import unittest
from PyFin.Math.RootFinder import Converged
from PyFin.Math.RootFinder import BracketedBrent
from PyFin.Math.RootFinder import Brent


def root_func(x):
    return x * x - 1.


class TestBrent(unittest.TestCase):

    def testBracketedBrent(self):
        converge = Converged(1e-8, 1e-8)

        low = 0., root_func(0.)
        high = 2., root_func(2.)

        brent = BracketedBrent(1e-8, low, high)
        iterations = 0

        while True and iterations < 100:
            iterations += 1
            x = brent.nextX()
            if converge.check_converge(brent, root_func(x)):
                break

        self.assertAlmostEqual(x, 1.)

    def testBracketedBrentNotBracketed(self):
        low = 0., 1.
        high = 2., 1.

        with self.assertRaises(ValueError):
            _ = BracketedBrent(1e-8, low, high)

    def testBrent(self):
        converge = Converged(1e-8, 1e-8)

        brent = Brent(1.5)
        iterations = 0

        while True and iterations < 100:
            iterations += 1
            x = brent.nextX()
            if converge.check_converge(brent, root_func(x)):
                break

        self.assertAlmostEqual(x, 1.)

        brent = Brent(-1.5)
        iterations = 0

        while True and iterations < 100:
            iterations += 1
            x = brent.nextX()
            if converge.check_converge(brent, root_func(x)):
                break

        self.assertAlmostEqual(x, -1.)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-13

@author: cheng.li
"""

import unittest
import copy
import tempfile
import pickle
import os
from PyFin.DateUtilities import Period
from PyFin.Enums import TimeUnits


class TestPeriod(unittest.TestCase):
    def testBasicArithmic(self):
        # test bad normalize
        testPriod = Period(length=1, units=TimeUnits.Years)
        testPriod._units = 10
        with self.assertRaises(TypeError):
            testPriod.normalize()

        # test plus method
        p1 = Period(length=0, units=TimeUnits.Days)
        p2 = Period(length=10, units=TimeUnits.Months)
        calculated = p1 + p2
        self.assertEqual(p2, calculated, "added value {0} should be equal to {1}".format(calculated, p2))

        p1 = Period(length=2, units=TimeUnits.Years)
        p2 = Period(length=13, units=TimeUnits.Months)
        calculated = p1 + p2
        expected = Period(length=37, units=TimeUnits.Months)
        self.assertEqual(expected, calculated, "added value {0} should be equal to {1}".format(calculated, expected))

        p2 = Period(length=2, units=TimeUnits.Weeks)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.BDays)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Days)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2._units = 10
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p1 = Period(length=13, units=TimeUnits.Months)
        p2 = Period(length=2, units=TimeUnits.Years)
        calculated = p1 + p2
        expected = Period(length=37, units=TimeUnits.Months)
        self.assertEqual(expected, calculated, "added value {0} should be equal to {1}".format(calculated, expected))

        p2 = Period(length=2, units=TimeUnits.Weeks)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.BDays)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Days)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2._units = 10
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p1 = Period(length=2, units=TimeUnits.Weeks)
        p2 = Period(length=7, units=TimeUnits.Days)
        calculated = p1 + p2
        expected = Period(length=21, units=TimeUnits.Days)
        self.assertEqual(expected, calculated, "added value {0} should be equal to {1}".format(calculated, expected))

        p2 = Period(length=2, units=TimeUnits.Months)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.BDays)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Years)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2._units = 10
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p1 = Period(length=7, units=TimeUnits.Days)
        p2 = Period(length=2, units=TimeUnits.Weeks)
        calculated = p1 + p2
        expected = Period(length=21, units=TimeUnits.Days)
        self.assertEqual(expected, calculated, "added value {0} should be equal to {1}".format(calculated, expected))

        p2 = Period(length=2, units=TimeUnits.Months)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.BDays)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Years)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2._units = 10
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p1 = Period(length=7, units=TimeUnits.BDays)

        p2 = Period(length=2, units=TimeUnits.Months)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Days)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Weeks)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.Years)
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2._units = 10
        with self.assertRaises(ValueError):
            _ = p1 + p2

        p2 = Period(length=2, units=TimeUnits.BDays)
        self.assertEqual(p1 + p2, Period('9B'))

        # test negative operator
        p1 = Period(length=-13, units=TimeUnits.Weeks)
        p2 = -p1
        self.assertEqual(p2, Period(length=13, units=TimeUnits.Weeks))

        # test less operator
        p1 = Period(length=0, units=TimeUnits.Days)
        p2 = Period(length=-3, units=TimeUnits.BDays)
        self.assertTrue(p2 < p1)

        # test sub operator
        p1 = Period(length=0, units=TimeUnits.Days)
        p2 = Period(length=-3, units=TimeUnits.BDays)
        self.assertEqual(p1 - p2, Period('3b'))

        # test string representation
        p1 = Period(length=12, units=TimeUnits.Months)
        self.assertEqual("1Y", p1.__str__())

    def testComparingOperators(self):
        p1 = Period(length=0, units=TimeUnits.Days)
        p2 = Period(length=1, units=TimeUnits.Days)
        self.assertTrue(p1 < p2)

        p1 = Period(length=13, units=TimeUnits.Months)
        p2 = Period(length=1, units=TimeUnits.Years)
        self.assertTrue(not p1 < p2)

        p1 = Period(length=1, units=TimeUnits.Years)
        p2 = Period(length=13, units=TimeUnits.Months)
        self.assertTrue(p1 < p2)

        p1 = Period(length=13, units=TimeUnits.Days)
        p2 = Period(length=2, units=TimeUnits.Weeks)
        self.assertTrue(p1 < p2)

        p1 = Period(length=2, units=TimeUnits.Weeks)
        p2 = Period(length=13, units=TimeUnits.Days)
        self.assertTrue(not p1 < p2)

        p1 = Period(length=1, units=TimeUnits.Years)
        p2 = Period(length=56, units=TimeUnits.Weeks)
        self.assertTrue(p1 < p2)

        p1 = Period(length=56, units=TimeUnits.Weeks)
        p2 = Period(length=1, units=TimeUnits.Years)
        self.assertTrue(not p1 < p2)

        p1 = Period(length=21, units=TimeUnits.Weeks)
        p2 = Period(length=5, units=TimeUnits.Months)

        with self.assertRaises(ValueError):
            _ = p1 < p2

        p1 = Period(length=21, units=TimeUnits.BDays)
        with self.assertRaises(ValueError):
            _ = p1 < p2

        # test not equal operator
        p1 = Period(length=1, units=TimeUnits.Days)
        p2 = Period(length=1, units=TimeUnits.Days)
        self.assertTrue(not p1 != p2)

        p2 = Period(length=1, units=TimeUnits.Years)
        self.assertTrue(p1 != p2)

        # test greater than operator
        p1 = Period(length=1, units=TimeUnits.Days)
        p2 = Period(length=2, units=TimeUnits.Days)
        self.assertEqual(p1 < p2, not p1 > p2)

    def testYearsMonthsAlgebra(self):
        oneYear = Period(length=1, units=TimeUnits.Years)
        sixMonths = Period(length=6, units=TimeUnits.Months)
        threeMonths = Period(length=3, units=TimeUnits.Months)

        n = 4
        flag = oneYear / n == threeMonths
        self.assertTrue(flag, "division error: {0} / {1:d}"
                              " not equal to {2}".format(oneYear, n, threeMonths))

        n = 2
        flag = oneYear / n == sixMonths
        self.assertTrue(flag, "division error: {0} / {1:d}"
                              " not equal to {2}".format(oneYear, n, sixMonths))

        sum = threeMonths
        sum += sixMonths
        flag = sum == Period(length=9, units=TimeUnits.Months)
        self.assertTrue(flag, "sum error: {0}"
                              " + {1}"
                              " != {2}".format(threeMonths, sixMonths, Period(length=9, units=TimeUnits.Months)))

        sum += oneYear
        flag = sum == Period(length=21, units=TimeUnits.Months)
        self.assertTrue(flag, "sum error: {0}"
                              " + {1}"
                              " + {2}"
                              " != {3}".format(threeMonths, sixMonths, oneYear, Period(length=21, units=TimeUnits.Months)))

        twelveMonths = Period(length=12, units=TimeUnits.Months)
        flag = twelveMonths.length() == 12
        self.assertTrue(flag, "normalization error: TwelveMonths.length"
                              " is {0:d}"
                              " instead of 12".format(twelveMonths.length()))
        flag = twelveMonths.units() == TimeUnits.Months
        self.assertTrue(flag, "normalization error: TwelveMonths.units"
                              " is {0:d}"
                              " instead of {1:d}".format(twelveMonths.units(), TimeUnits.Months))

        normalizedTwelveMonths = Period(length=12, units=TimeUnits.Months)
        normalizedTwelveMonths = normalizedTwelveMonths.normalize()
        flag = normalizedTwelveMonths.length() == 1
        self.assertTrue(flag, "normalization error: TwelveMonths.length"
                              " is {0:d}"
                              " instead of 1".format(twelveMonths.length()))
        flag = normalizedTwelveMonths.units() == TimeUnits.Years
        self.assertTrue(flag, "normalization error: TwelveMonths.units"
                              " is {0:d}"
                              " instead of {1:d}".format(twelveMonths.units(), TimeUnits.Years))

        thirtyDays = Period(length=30, units=TimeUnits.Days)
        normalizedThirtyDays = thirtyDays.normalize()
        flag = normalizedThirtyDays.units() == TimeUnits.Days
        self.assertTrue(flag, "normalization error: ThirtyDays.units"
                              " is {0:d}"
                              " instead of {1:d}".format(normalizedThirtyDays.units(), TimeUnits.Days))

        thirtyBDays = Period(length=30, units=TimeUnits.BDays)
        normalizedThirtyBDays = thirtyBDays.normalize()
        flag = normalizedThirtyBDays.units() == TimeUnits.BDays
        self.assertTrue(flag, "normalization error: ThirtyBDays.units"
                              " is {0:d}"
                              " instead of {1:d}".format(normalizedThirtyBDays.units(), TimeUnits.BDays))

    def testWeeksDaysAlgebra(self):
        twoWeeks = Period(length=2, units=TimeUnits.Weeks)
        oneWeek = Period(length=1, units=TimeUnits.Weeks)
        threeDays = Period(length=3, units=TimeUnits.Days)
        oneDay = Period(length=1, units=TimeUnits.Days)

        n = 2
        flag = twoWeeks / n == oneWeek
        self.assertTrue(flag, "division error: {0} / {1:d}"
                              " not equal to {2}".format(twoWeeks, n, oneWeek))

        n = 7
        flag = oneWeek / 7 == oneDay
        self.assertTrue(flag, "division error: {0} / {1:d}"
                              " not equal to {2}".format(oneWeek, n, oneDay))

        sum = threeDays
        sum += oneDay
        flag = sum == Period(length=4, units=TimeUnits.Days)
        self.assertTrue(flag, "sum error: {0}"
                              " + {1}"
                              " != {2}".format(threeDays, oneDay, Period(length=4, units=TimeUnits.Days)))

        sum += oneWeek
        flag = sum == Period(length=11, units=TimeUnits.Days)
        self.assertTrue(flag, "sum error: {0}"
                              " + {1}"
                              " + {2}"
                              " != {3}".format(threeDays, oneDay, oneWeek, Period(length=11, units=TimeUnits.Days)))

        sevenDays = Period(length=7, units=TimeUnits.Days)
        flag = sevenDays.length() == 7
        self.assertTrue(flag, "normalization error: sevenDays.length"
                              " is {0:d}"
                              " instead of 7".format(sevenDays.length()))
        flag = sevenDays.units() == TimeUnits.Days
        self.assertTrue(flag, "normalization error: sevenDays.units"
                              " is {0:d}"
                              " instead of {1:d}".format(sevenDays.units(), TimeUnits.Days))

        normalizedSevenDays = sevenDays.normalize()
        flag = normalizedSevenDays.length() == 1
        self.assertTrue(flag, "normalization error: normalizedSevenDays.length"
                              " is {0:d}"
                              " instead of 1".format(normalizedSevenDays.length()))
        flag = normalizedSevenDays.units() == TimeUnits.Weeks
        self.assertTrue(flag, "normalization error: TwelveMonths.units"
                              " is {0:d}"
                              " instead of {1:d}".format(normalizedSevenDays.units(), TimeUnits.Weeks))

    def testPeriodDeepCopy(self):

        p1 = Period('36m')
        p2 = copy.deepcopy(p1)

        self.assertEqual(p1, p2)

    def testPeriodPickle(self):
        p1 = Period('36m')

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(p1, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickled_period = pickle.load(f2)
            self.assertEqual(p1, pickled_period)

        os.unlink(f.name)



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-7

@author: cheng.li
"""

import xlwings as xw
import numpy as np
from PyFin.PricingEngines import sabrVolatility as sbv
from PyFin.PricingEngines import sabrVolatilities as sbvs
from PyFin.PricingEngines import sabrCalibration as scb
from PyFin.PricingEngines import sviVolatility as svi
from PyFin.PricingEngines import sviVolatilities as svis
from PyFin.PricingEngines import sviCalibration as svic


@xw.func
def sabrVolatility(strike, forward, expiry, alpha, beta, nu, rho):
    return sbv(strike, forward, expiry, alpha, beta, nu, rho)


@xw.func
@xw.ret(expand='table')
@xw.arg('strikes', np.array, dim=1)
def sabrVolatilities(strikes, forward, expiry, alpha, beta, nu, rho):
    return sbvs(strikes, forward, expiry, alpha, beta, nu, rho)


@xw.func
@xw.ret(expand='table')
@xw.arg('strikes', np.array, dim=1)
@xw.arg('volatilites', np.array, dim=1)
def sabrCalibration(strikes,
                    volatilites,
                    forward,
                    expiryTime,
                    intialAlpha,
                    initialBeta,
                    initialNu,
                    initialRho,
                    isFixedAlpha=False,
                    isFixedBeta=False,
                    isFixedNu=False,
                    isFixedRho=False,
                    method='trf'):
    x = scb(strikes,
            volatilites,
            forward,
            expiryTime,
            intialAlpha,
            initialBeta,
            initialNu,
            initialRho,
            isFixedAlpha,
            isFixedBeta,
            isFixedNu,
            isFixedRho,
            method)
    return x[0]


@xw.func
def sviVolatility(strike, forward, expiry, a, b, sigma, rho, m):
    return svi(strike, forward, expiry, a, b, sigma, rho, m)


@xw.func
@xw.ret(expand='table')
@xw.arg('strikes', np.array, dim=1)
def sviVolatilities(strikes, forward, expiry, a, b, sigma, rho, m):
    return svis(strikes, forward, expiry, a, b, sigma, rho, m)


@xw.func
@xw.ret(expand='table')
@xw.arg('strikes', np.array, dim=1)
@xw.arg('volatilites', np.array, dim=1)
def sviCalibration(strikes,
                   volatilites,
                   forward,
                   expiryTime,
                   initialA,
                   initialB,
                   initialSigma,
                   initialRho,
                   initialM,
                   isFixedA=False,
                   isFixedB=False,
                   isFixedSigma=False,
                   isFixedRho=False,
                   isFixedM=False,
                   method='trf'):
    x = svic(strikes,
             volatilites,
             forward,
             expiryTime,
             initialA,
             initialB,
             initialSigma,
             initialRho,
             initialM,
             isFixedA,
             isFixedB,
             isFixedSigma,
             isFixedRho,
             isFixedM,
             method)
    return x[0]


# if __name__ == '__main__':
#     xw.serve()


# -*- coding: utf-8 -*-
u"""
Created on 2015-8-12

@author: cheng.li
"""

import unittest
import copy
import pickle
import tempfile
import os
import numpy as np
import pandas as pd
from PyFin.Enums import Factors
from PyFin.Analysis.SeriesValues import SeriesValues
from PyFin.Analysis.SecurityValueHolders import FilteredSecurityValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityCurrentValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityIIFValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityShiftedValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from PyFin.Analysis.SecurityValueHolders import SecurityConstArrayValueHolder
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingAverage
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingMax
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingMin
from PyFin.Analysis.TechnicalAnalysis import SecurityMovingSum


class TestSecurityValueHolders(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        sample1 = np.random.randn(1000, 2)
        sample2 = np.random.randn(1000, 2)

        self.datas = {'aapl': {'close': sample1[:, 0], 'open': sample1[:, 1]},
                      'ibm': {'close': sample2[:, 0], 'open': sample2[:, 1]}}

        def check_values(expected, calculated):
            for name in calculated.index():
                self.assertEqual(calculated[name], expected[name], "for the name {0}\n"
                                                                   "expected:   {1}\n"
                                                                   "calculated: {2}".format(name,
                                                                                            expected[name],
                                                                                            calculated[name]))

        self.checker = check_values

    def testSecurityValueHolderWithGetItemUseInt(self):

        shift = 2
        benchmark = SecurityShiftedValueHolder(shift, SecurityLatestValueHolder(x='close'))
        testValueHolder = SecurityLatestValueHolder(x='close')[shift]

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            testValueHolder.push(data)

            if i < shift:
                continue

            calculated = testValueHolder.value
            expected = benchmark.value

            self.assertEqual(calculated.name_mapping, expected.name_mapping)
            np.testing.assert_array_almost_equal(calculated.values, expected.values)

    def testSecurityWhereValueHolder(self):
        benchmark = SecurityLatestValueHolder(x='close')

        testValueHolder = SecurityIIFValueHolder(benchmark > 0, benchmark, -benchmark)
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            testValueHolder.push(data)

            calculated = testValueHolder.value

            self.assertTrue(np.all((calculated > 0.).values))

            for name in calculated.index():
                rawValue = data[name][Factors.CLOSE]
                if rawValue > 0.:
                    self.assertAlmostEqual(rawValue, calculated[name])
                else:
                    self.assertAlmostEqual(-rawValue, calculated[name])

    def testSecurityConstArrayValueHolder(self):
        constArray = SecurityConstArrayValueHolder({'a': 1.0, 'b': 2.0})

        testValue = constArray.value
        expectedValue = SeriesValues(np.array([1.0, 2.0]), index=['a', 'b'])

        for name in expectedValue.index():
            self.assertAlmostEqual(testValue[name], expectedValue[name])

        testValue = constArray.value_by_name('a')
        self.assertAlmostEqual(1.0, testValue)

        testValue = constArray.value_by_names(['b', 'a', 'c'])
        expectedValue = SeriesValues(np.array([2.0, 1.0, np.nan]), index=['b', 'a', 'c'])

        np.testing.assert_array_almost_equal(testValue.values, expectedValue.values)
        self.assertEqual(testValue.name_mapping, expectedValue.name_mapping)

    def testSecurityConstArrayValueHolderWorkWithOthers(self):
        dictValue = {'aapl': 1.0, 'ibm': 2.0}
        constArray = SecurityConstArrayValueHolder(dictValue)

        benchmark = SecurityLatestValueHolder(x='close')
        multiplied = benchmark * constArray

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            constArray.push(data)
            multiplied.push(data)

            testValue = benchmark.value
            expectedValue = multiplied.value

            for name in expectedValue.index():
                self.assertAlmostEqual(testValue[name] * dictValue[name], expectedValue[name])

    def testSecurityAndOperatorValueHolder(self):
        benchmark = SecurityLatestValueHolder(x='close') > 0
        benchmark2 = SecurityLatestValueHolder(x='open') > 0

        testValueHolder = benchmark & benchmark2

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            testValueHolder.push(data)
            calculated = testValueHolder.value

            for name in calculated.index():
                close_value = data[name][Factors.CLOSE]
                open_value = data[name][Factors.OPEN]

                if close_value > 0. and open_value > 0:
                    self.assertTrue(calculated[name])
                else:
                    self.assertFalse(calculated[name])

    def testSecurityOrOperatorValueHolder(self):
        benchmark = SecurityLatestValueHolder(x='close') > 0
        benchmark2 = SecurityLatestValueHolder(x='open') > 0

        testValueHolder = benchmark | benchmark2

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            testValueHolder.push(data)
            calculated = testValueHolder.value

            for name in calculated.index():
                close_value = data[name][Factors.CLOSE]
                open_value = data[name][Factors.OPEN]

                if close_value > 0. or open_value > 0:
                    self.assertTrue(calculated[name])
                else:
                    self.assertFalse(calculated[name])

    def testSecurityXorOperatorValueHolder(self):
        benchmark = SecurityLatestValueHolder(x='close') > 0
        benchmark2 = SecurityLatestValueHolder(x='open') > 0

        testValueHolder = benchmark ^ benchmark2

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            testValueHolder.push(data)
            calculated = testValueHolder.value

            for name in calculated.index():
                close_value = data[name][Factors.CLOSE]
                open_value = data[name][Factors.OPEN]

                self.assertEqual(calculated[name][0], close_value > 0)
                self.assertEqual(calculated[name][1], open_value > 0)

    def testSecurityWhereValueHolderWithSymbolName(self):
        benchmark = SecurityLatestValueHolder(x='close')

        testValueHolder = SecurityIIFValueHolder(benchmark > 0, 'open', -benchmark)
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            testValueHolder.push(data)

            calculated = testValueHolder.value

            for name in calculated.index():
                rawValue = data[name][Factors.CLOSE]
                if rawValue > 0.:
                    self.assertAlmostEqual(data[name][Factors.OPEN], calculated[name])
                else:
                    self.assertAlmostEqual(-rawValue, calculated[name])

        testValueHolder = SecurityIIFValueHolder(benchmark > 0, 'open', 'close')
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            testValueHolder.push(data)

            calculated = testValueHolder.value

            for name in calculated.index():
                rawValue = data[name][Factors.CLOSE]
                if rawValue > 0.:
                    self.assertAlmostEqual(data[name][Factors.OPEN], calculated[name])
                else:
                    self.assertAlmostEqual(rawValue, calculated[name])

    def testSecurityWhereValueHolderWithMissingSymbol(self):
        benchmark = SecurityLatestValueHolder(x='open')
        benchmark2 = SecurityLatestValueHolder(x='close')
        testValueHolder = SecurityIIFValueHolder(benchmark < 0, 'open', -benchmark2)

        data = {'aapl': {'close': 2.0, 'open': 1.5},
                'ibm': {'open': 1.7}}

        benchmark.push(data)
        benchmark2.push(data)
        testValueHolder.push(data)

        calculated = testValueHolder.value

        self.assertAlmostEqual(calculated['aapl'], -data['aapl']['close'])
        self.assertTrue(np.isnan(calculated['ibm']))

    def testSecurityWhereValueHolderWithScalarValue(self):
        benchmark = SecurityLatestValueHolder(x='open')
        testValueHolder = SecurityIIFValueHolder(benchmark > 0, 1., -1.)
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            testValueHolder.push(data)

            calculated = testValueHolder.value

            for name in calculated.index():
                rawValue = data[name][Factors.OPEN]
                if rawValue > 0.:
                    self.assertAlmostEqual(1., calculated[name])
                    self.assertAlmostEqual(1., testValueHolder.value_by_name(name))
                else:
                    self.assertAlmostEqual(-1., calculated[name])
                    self.assertAlmostEqual(-1., testValueHolder.value_by_name(name))

    def testFilteredSecurityValueHolder(self):
        benchmark = SecurityLatestValueHolder(x='close') > 0
        filtered = FilteredSecurityValueHolder(benchmark, benchmark)

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            filtered.push(data)
            try:
                _ = filtered['aapl']
                self.assertTrue(benchmark['aapl'])
            except KeyError:
                self.assertFalse(benchmark['aapl'])

            try:
                _ = filtered['ibm']
                self.assertTrue(benchmark['ibm'])
            except KeyError:
                self.assertFalse(benchmark['ibm'])

    def testFilteredSecurityValueHolderWithGetItemMethod(self):
        benchmark = SecurityLatestValueHolder(x='close') > 0
        filtered = benchmark[benchmark]

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            benchmark.push(data)
            filtered.push(data)

            try:
                _ = filtered['aapl']
                self.assertTrue(benchmark['aapl'])
            except KeyError:
                self.assertFalse(benchmark['aapl'])

            try:
                _ = filtered['ibm']
                self.assertTrue(benchmark['ibm'])
            except KeyError:
                self.assertFalse(benchmark['ibm'])

    def testFilteredSecurityValueHolderTransform(self):
        rawHolder = SecurityLatestValueHolder(x='close')
        fliterFlag = rawHolder >= 0
        filteredHolder = rawHolder[fliterFlag]

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            rawHolder.push(data)
            filteredHolder.push(data)

            filteredValues = filteredHolder.value >= 0.

            self.assertTrue(np.all(filteredValues.values))

            rawValues = rawHolder.value
            filteredValues = filteredHolder.value
            for name in rawValues.index():
                if rawValues[name] >= 0.:
                    self.assertTrue(name in filteredValues)

    def testSecurityValueHolderIsFull(self):
        test = SecurityMovingMax(2, x='close')

        data = {'aapl': {'close': 1.0},
                'ibm': {'close': 2.0}}
        test.push(data)
        self.assertEqual(test.isFull, False)

        data = {'aapl': {'close': 1.0},
                'ibm': {'close': 13.0}}
        test.push(data)
        self.assertEqual(test.isFull, True)

    def testSecuritiesValuesComparison(self):

        benchmarkValues = SeriesValues({'AAPL': 1.0, 'IBM': 2.0, 'GOOG': 3.0})
        calculated = benchmarkValues > 1.5
        expected = SeriesValues({'AAPL': False, 'IBM': True, 'GOOG': True})
        self.checker(expected, calculated)

        calculated = benchmarkValues < 1.5
        expected = SeriesValues({'AAPL': True, 'IBM': False, 'GOOG': False})
        self.checker(expected, calculated)

        calculated = benchmarkValues >= 1.0
        expected = SeriesValues({'AAPL': True, 'IBM': True, 'GOOG': True})
        self.checker(expected, calculated)

        calculated = benchmarkValues <= 2.0
        expected = SeriesValues({'AAPL': True, 'IBM': True, 'GOOG': False})
        self.checker(expected, calculated)

        benchmarkValues = SeriesValues({'AAPL': False, 'IBM': True, 'GOOG': False})
        calculated = benchmarkValues & True
        expected = SeriesValues({'AAPL': False, 'IBM': True, 'GOOG': False})
        self.checker(expected, calculated)

        calculated = True & benchmarkValues
        expected = SeriesValues({'AAPL': False, 'IBM': True, 'GOOG': False})
        self.checker(expected, calculated)

        calculated = benchmarkValues | True
        expected = SeriesValues({'AAPL': True, 'IBM': True, 'GOOG': True})
        self.checker(expected, calculated)

        calculated = True | benchmarkValues
        expected = SeriesValues({'AAPL': True, 'IBM': True, 'GOOG': True})
        self.checker(expected, calculated)

        benchmarkValues = SeriesValues({'AAPL': 1.0, 'IBM': 2.0, 'GOOG': 3.0})
        calculated = benchmarkValues == 2.0
        expected = SeriesValues({'AAPL': False, 'IBM': True, 'GOOG': False})
        self.checker(expected, calculated)

        calculated = benchmarkValues != 2.0
        expected = SeriesValues({'AAPL': True, 'IBM': False, 'GOOG': True})
        self.checker(expected, calculated)

    def testSecuritiesValuesArithmetic(self):
        benchmarkValues = {'AAPL': 1.0, 'IBM': 2.0, 'GOOG': 3.0}
        values = SeriesValues(benchmarkValues)

        self.assertEqual(len(benchmarkValues), len(values))

        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key], values[key], 12)

        negValues = -values
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key], -negValues[key], 12)

        benchmarkValues2 = {'AAPL': 3.0, 'IBM': 2.0, 'GOOG': 1.0}
        values2 = SeriesValues(benchmarkValues2)
        addedValue = values + values2
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] + benchmarkValues2[key], addedValue[key], 12)

        subbedValues = values - values2
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] - benchmarkValues2[key], subbedValues[key], 12)

        multiValues = values * values2
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] * benchmarkValues2[key], multiValues[key], 12)

        divValues = values / values2
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] / benchmarkValues2[key], divValues[key], 12)

        # check operated with scalar
        addedValue = values + 2.0
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] + 2.0, addedValue[key], 12)

        subbedValue = values - 2.0
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] - 2.0, subbedValue[key], 12)

        multiValues = values * 2.0
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] * 2.0, multiValues[key], 12)

        divValues = values / 2.0
        for key in benchmarkValues:
            self.assertAlmostEqual(benchmarkValues[key] / 2.0, divValues[key], 12)

        # check right associate operators
        addedValue = 2.0 + values
        for key in benchmarkValues:
            self.assertAlmostEqual(2.0 + benchmarkValues[key], addedValue[key], 12)

        subbedValue = 2.0 - values
        for key in benchmarkValues:
            self.assertAlmostEqual(2.0 - benchmarkValues[key], subbedValue[key], 12)

        multiValues = 2.0 * values
        for key in benchmarkValues:
            self.assertAlmostEqual(2.0 * benchmarkValues[key], multiValues[key], 12)

        divValues = 2.0 / values
        for key in benchmarkValues:
            self.assertAlmostEqual(2.0 / benchmarkValues[key], divValues[key], 12)

    def testBasicFunctions(self):
        window = 10
        pNames = 'close'
        symbolList = ['aapl', 'ibm']
        testValueHolder = SecurityMovingAverage(window, pNames)

        testValueHolder.push({'aapl': {'close': 1.0}, 'ibm': {'close': 2.0}})

        self.assertEqual(set(testValueHolder.symbolList), set(symbolList))
        self.assertEqual(testValueHolder.window, window)

        # test binary operated value holder
        window2 = 5
        pNames2 = 'open'
        test2 = SecurityMovingMax(window2, pNames2)
        binaryValueHolder = testValueHolder + test2

        self.assertEqual(binaryValueHolder.window, max(window, window2))

        # test compounded operated value holder
        test3 = SecurityMovingMax(window2, testValueHolder)
        self.assertEqual(test3.window, window + window2)

        # test compounded twice
        test4 = SecurityMovingMax(window2, test3)
        self.assertEqual(test4.window, window + 2 * window2)

    def testItemizedValueHolder(self):
        window = 10
        pNames = 'close'
        test = SecurityMovingAverage(window, pNames)
        test.push({'aapl': {'close': 10.0}, 'ibm': {'close': 15.0}, 'goog': {'close': 17.0}})
        test.push({'aapl': {'close': 12.0}, 'ibm': {'close': 10.0}, 'goog': {'close': 13.0}})

        expected = {'ibm': 12.5, 'goog': 15.0}
        for s in expected:
            self.assertAlmostEqual(test[s], expected[s])

    def testAddedSecurityValueHolders(self):
        window1 = 10
        window2 = 5
        dependency1 = Factors.CLOSE
        dependency2 = Factors.OPEN
        ma = SecurityMovingAverage(window1, dependency1)
        mm = SecurityMovingSum(window2, dependency2)
        combined = ma + mm

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            mm.push(data)
            combined.push(data)

            expected = ma.value + mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testAddedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        mm = SecurityMovingSum(window, dependency)
        combined = 2.0 + mm
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            mm.push(data)
            combined.push(data)

            expected = 2.0 + mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testAddedSecurityValueHolderWithName(self):
        window = 10
        dependency = 'close'
        mm = SecurityMovingSum(window, dependency)
        combined = 'open' + mm

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            mm.push(data)
            combined.push(data)

            base = mm.value
            calculated = combined.value
            for name in calculated.index():
                self.assertAlmostEqual(base[name] + data[name]['open'], calculated[name], 12)

    def testRAddedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        ma = SecurityMovingAverage(window, dependency)
        combined = ma + 2.0
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            combined.push(data)

            expected = ma.value + 2.0
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testSubbedSecurityValueHolder(self):
        window1 = 10
        window2 = 5
        dependency1 = Factors.CLOSE
        dependency2 = Factors.OPEN
        ma = SecurityMovingAverage(window1, dependency1)
        mm = SecurityMovingSum(window2, dependency2)
        combined = ma - mm

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            mm.push(data)
            combined.push(data)

            expected = ma.value - mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testSubbedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        mm = SecurityMovingSum(window, dependency)
        combined = 2.0 - mm
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            mm.push(data)
            combined.push(data)

            expected = 2.0 - mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testRSubbedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        ma = SecurityMovingAverage(window, dependency)
        combined = ma - 2.0
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            combined.push(data)

            expected = ma.value - 2.0
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testMultipliedSecurityValueHolders(self):
        window1 = 10
        window2 = 5
        dependency1 = Factors.CLOSE
        dependency2 = Factors.OPEN
        ma = SecurityMovingAverage(window1, dependency1)
        mm = SecurityMovingSum(window2, dependency2)
        combined = ma * mm

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            mm.push(data)
            combined.push(data)

            expected = ma.value * mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testMultipliedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        mm = SecurityMovingSum(window, dependency)
        combined = 2.0 * mm
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            mm.push(data)
            combined.push(data)

            expected = 2.0 * mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testRMultipliedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        ma = SecurityMovingAverage(window, dependency)
        combined = ma * 2.0
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            combined.push(data)

            expected = ma.value * 2.0
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testDividedSecurityValueHolders(self):
        window1 = 10
        window2 = 5
        dependency1 = Factors.CLOSE
        dependency2 = Factors.OPEN
        ma = SecurityMovingAverage(window1, dependency1)
        mm = SecurityMovingSum(window2, dependency2)
        combined = ma / mm

        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            mm.push(data)
            combined.push(data)

            expected = ma.value / mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testDividedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        mm = SecurityMovingSum(window, dependency)
        combined = 2.0 / mm
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            mm.push(data)
            combined.push(data)

            expected = 2.0 / mm.value
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testRDividedSecurityValueHoldersWithScalar(self):
        window = 10
        dependency = 'close'
        ma = SecurityMovingAverage(window, dependency)
        combined = ma / 2.0
        for i in range(len(self.datas['aapl']['close'])):
            data = {'aapl': {Factors.CLOSE: self.datas['aapl'][Factors.CLOSE][i],
                             Factors.OPEN: self.datas['aapl'][Factors.OPEN][i]},
                    'ibm': {Factors.CLOSE: self.datas['ibm'][Factors.CLOSE][i],
                            Factors.OPEN: self.datas['ibm'][Factors.OPEN][i]}}
            ma.push(data)
            combined.push(data)

            expected = ma.value / 2.0
            calculated = combined.value
            for name in expected.index():
                self.assertAlmostEqual(expected[name], calculated[name], 12)

    def testCombinedSecurityValueHolderWithoutData(self):
        ma = SecurityMovingAverage(10, 'close')
        calculated = ma.value

        for name in calculated.index():
            self.assertTrue(np.isnan(calculated[name]))

    def testShiftedSecurityValueHolder(self):
        mm = SecurityMovingAverage(2, 'close')
        shifted1 = mm.shift(1)

        data1 = {'aapl': {'close': 1.0},
                 'ibm': {'close': 2.0},
                 'goog': {'close': 3.0}}
        shifted1.push(data1)
        calculated = shifted1.value
        for name in calculated.index():
            self.assertTrue(np.isnan(calculated[name]))

        data2 = {'aapl': {'close': 2.0},
                 'ibm': {'close': 3.0},
                 'goog': {'close': 4.0}}

        shifted1.push(data2)
        expected = SeriesValues({'aapl': 1.0,
                                 'ibm': 2.0,
                                 'goog': 3.0})
        calculated = shifted1.value
        for name in expected.index():
            self.assertAlmostEqual(expected[name], calculated[name])

        data3 = ({'aapl': {'close': 3.0},
                  'ibm': {'close': 4.0},
                  'goog': {'close': 5.0}})

        shifted1.push(data3)
        expected = SeriesValues({'aapl': 1.5,
                                 'ibm': 2.5,
                                 'goog': 3.5})
        calculated = shifted1.value
        for name in expected.index():
            self.assertAlmostEqual(expected[name], calculated[name])

    def testDeltaSecurityValueHolder(self):
        mm = SecurityMovingAverage(2, 'close')
        delta1 = SecurityDeltaValueHolder(2, mm)

        data1 = {'aapl': {'close': 1.0},
                 'ibm': {'close': 2.0},
                 'goog': {'close': 3.0}}
        delta1.push(data1)
        calculated = delta1.value
        for name in calculated.index():
            self.assertTrue(np.isnan(calculated[name]))

        data2 = {'aapl': {'close': 2.0},
                 'ibm': {'close': 3.0},
                 'goog': {'close': 4.0}}

        delta1.push(data2)

        data3 = {'aapl': {'close': 3.0},
                 'ibm': {'close': 4.0},
                 'goog': {'close': 5.0}}

        delta1.push(data3)

        expected = SeriesValues({'aapl': 1.5,
                                 'ibm': 1.5,
                                 'goog': 1.5})
        calculated = delta1.value
        for name in expected.index():
            self.assertAlmostEqual(expected[name], calculated[name])

        data4 = ({'aapl': {'close': 4.0},
                  'ibm': {'close': 5.0},
                  'goog': {'close': 6.0}})

        delta1.push(data4)
        expected = SeriesValues({'aapl': 2.0,
                                 'ibm': 2.0,
                                 'goog': 2.0})
        calculated = delta1.value
        for name in expected.index():
            self.assertAlmostEqual(expected[name], calculated[name])

    def testShiftedSecurityValueHolderWithLengthZero(self):
        mm = SecurityMovingAverage(2, 'close')
        with self.assertRaises(ValueError):
            _ = mm.shift(0)

    def testShiftedSecurityValueHolderDeepcopy(self):
        tested = SecurityShiftedValueHolder(2, SecurityLatestValueHolder('close'))

        data = dict(aapl=dict(close=1.0),
                    ibm=dict(close=2.0))
        tested.push(data)

        data = dict(aapl=dict(close=3.0),
                    ibm=dict(close=4.0))
        tested.push(data)

        data = dict(aapl=dict(close=5.0),
                    ibm=dict(close=6.0))
        tested.push(data)

        copied = copy.deepcopy(tested)

        expected = tested.value
        calcualted = copied.value
        for name in expected.index():
            self.assertAlmostEqual(expected[name], calcualted[name])

    def testShiftedSecurityValueHolderPickle(self):
        tested = SecurityShiftedValueHolder(2, SecurityLatestValueHolder('close'))

        data = dict(aapl=dict(close=1.0),
                    ibm=dict(close=2.0))
        tested.push(data)

        data = dict(aapl=dict(close=3.0),
                    ibm=dict(close=4.0))
        tested.push(data)

        data = dict(aapl=dict(close=5.0),
                    ibm=dict(close=6.0))
        tested.push(data)

        with tempfile.NamedTemporaryFile('w+b', delete=False) as f:
            pickle.dump(tested, f)

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)

        expected = tested.value
        calcualted = pickled.value
        for name in expected.index():
            self.assertAlmostEqual(expected[name], calcualted[name])

    def testLtSecurityValueHolder(self):
        filter = SecurityLatestValueHolder('close') < 10.0
        ma = FilteredSecurityValueHolder(SecurityMovingAverage(10, 'close'), filter)

        data = {'aapl': {'close': 15.},
                'ibm': {'close': 8.},
                'goog': {'close': 7.}}

        ma.push(data)
        expected = {'ibm': 8.0, 'goog': 7.0}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

        data = {'aapl': {'close': 9.},
                'ibm': {'close': 11.},
                'goog': {'close': 8.}}

        ma.push(data)
        expected = {'aapl': 12.0, 'goog': 7.5}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

    def testFilterSecurityValueHolderWorkWithStr(self):
        filter = SecurityLatestValueHolder('code') == 2
        ma = FilteredSecurityValueHolder(SecurityMovingAverage(10, 'close'), filter)

        data = {'aapl': {'code': 1, 'close': 15.},
                'ibm': {'code': 2, 'close': 10.},
                'goog': {'code': 3, 'close': 7.}}

        ma.push(data)
        expected = {'ibm': 10.0}
        calculated = ma.value

        for name in calculated.index():
            self.assertAlmostEqual(expected[name], calculated[name], 15)

    def testLeSecurityValueHolder(self):
        filter = SecurityLatestValueHolder('close') <= 10.0
        ma = FilteredSecurityValueHolder(SecurityMovingAverage(10, 'close'), filter)

        data = {'aapl': {'close': 15.},
                'ibm': {'close': 10.},
                'goog': {'close': 7.}}

        ma.push(data)
        expected = {'ibm': 10.0, 'goog': 7.0}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

        data = {'aapl': {'close': 10.},
                'ibm': {'close': 11.},
                'goog': {'close': 8.}}

        ma.push(data)
        expected = {'aapl': 12.5, 'goog': 7.5}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

    def testGtSecurityValueHolder(self):
        filter = SecurityLatestValueHolder('close') > 10.0
        ma = FilteredSecurityValueHolder(SecurityMovingAverage(10, 'close'), filter)

        data = {'aapl': {'close': 15.},
                'ibm': {'close': 8.},
                'goog': {'close': 7.}}

        ma.push(data)
        expected = {'aapl': 15.}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

        data = {'aapl': {'close': 9.},
                'ibm': {'close': 11.},
                'goog': {'close': 8.}}

        ma.push(data)
        expected = {'ibm': 9.5}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

    def testGeSecurityValueHolder(self):
        filter = SecurityMovingAverage(1, 'close') >= 10.0
        ma = FilteredSecurityValueHolder(SecurityMovingAverage(10, 'close'), filter)

        data = {'aapl': {'close': 15.},
                'ibm': {'close': 10.},
                'goog': {'close': 7.}}

        ma.push(data)
        expected = {'aapl': 15., 'ibm': 10.}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

        data = {'aapl': {'close': 10.},
                'ibm': {'close': 11.},
                'goog': {'close': 8.}}

        ma.push(data)
        expected = {'aapl': 12.5, 'ibm': 10.5}
        calculated = ma.value
        for name in expected:
            self.assertAlmostEqual(expected[name], calculated[name], 15)

    def testTransformWithCategory(self):
        test_df = pd.DataFrame({'code': [1, 2, 3, 4, 1, 2, 3],
                                'b': [4, 5, 6, 7, 6, 5, 4],
                                'c': [9, 8, 7, 6, 5, 4, 3]},
                               index=[1, 1, 1, 1, 2, 2, 2],
                               dtype=float)

        expression = SecurityMovingMax(20, 'b') + SecurityMovingMin(20, 'c')
        calculated = expression.transform(test_df, name='new_factor', category_field='code')
        expected = [13., 13., 13., 13., 11., 9, 9.]

        np.testing.assert_array_almost_equal(calculated['new_factor'], expected)

    def testTransformWithoutCategory(self):
        test_df = pd.DataFrame({'b': [4, 5, 6, 7, 6, 5, 4],
                                'c': [9, 8, 7, 6, 5, 4, 3]},
                               index=[1, 2, 3, 4, 5, 6, 7],
                               dtype=float)

        expression = SecurityMovingMax(20, 'b') + SecurityMovingMin(20, 'c')
        calculated = expression.transform(test_df, name='new_factor')
        expected = [13., 13., 13., 13., 12., 11., 10.]
        np.testing.assert_array_almost_equal(calculated['new_factor'], expected)

    def testTransformWithDifferentGroup(self):
        test_df = pd.DataFrame({'code': [1, 2, 3, 4, 5, 6, 7],
                                'b': [4, 5, 6, 7, 6, 5, 4],
                                'c': [9, 8, 7, 6, 5, 4, 3]},
                               index=[1, 1, 1, 1, 2, 2, 2],
                               dtype=float)

        expression = SecurityMovingAverage(2, 'b')
        calculated = expression.transform(test_df, name='new_factor', category_field='code')

        expected = [4., 5., 6., 7., 6., 5., 4.]
        np.testing.assert_array_almost_equal(calculated['new_factor'], expected)

    def testTransformWithoutCategoryAndDuplicatedIndex(self):
        sample_data = pd.DataFrame(
            data={'code': [1, 2, 1, 2, 1, 2],
                  'open': [2.0, 1.0, 1.5, 3.0, 2.4, 3.5],
                  'close': [1.7, 1.6, 0.9, 3.8, 1.6, 2.1]},
            index=[1, 1, 2, 2, 3, 3],
            dtype=float
        )

        expression = SecurityMovingAverage(2, 'close')
        calculated = expression.transform(sample_data, name='new_factor')

        expected = sample_data.rolling(window=2).mean()

        np.testing.assert_array_almost_equal(calculated['new_factor'].values[1:],
                                             expected['close'].values[1:])

    def testSecurityCurrentValueHolder(self):
        current = SecurityCurrentValueHolder('x')

        data1 = dict(aapl=dict(x=1),
                     ibm=dict(x=2))

        current.push(data1)
        calculated = current.value
        for name in data1.keys():
            self.assertAlmostEqual(data1[name]['x'], calculated[name])

        data2 = dict(aapl=dict(x=3),
                     ibm=dict(x=np.nan))
        current.push(data2)
        calculated = current.value
        for name in data1.keys():
            expected = data2[name]['x']
            if not np.isnan(expected):
                self.assertAlmostEqual(expected, calculated[name])
            else:
                self.assertTrue(np.isnan(calculated[name]))

    def testSecurityLatestValueHolder(self):
        latest = SecurityLatestValueHolder('x')

        data1 = dict(aapl=dict(x=1),
                     ibm=dict(x=2))

        latest.push(data1)
        calculated = latest.value
        for name in data1.keys():
            self.assertAlmostEqual(data1[name]['x'], calculated[name])

        data2 = dict(aapl=dict(x=3),
                     ibm=dict(x=np.nan))
        latest.push(data2)
        calculated = latest.value
        for name in data1.keys():
            expected = data2[name]['x']
            if not np.isnan(expected):
                self.assertAlmostEqual(expected, calculated[name])
            else:
                self.assertAlmostEqual(data1[name]['x'], calculated[name])

    def testSecurityLatestValueHolderDeepcopy(self):
        latest = SecurityLatestValueHolder('x')

        data = dict(aapl=dict(x=1),
                    ibm=dict(x=2))

        latest.push(data)

        copied = copy.deepcopy(latest)

        calculated = copied.value

        for name in data.keys():
            self.assertAlmostEqual(data[name]['x'], calculated[name])

    def testSecurityLatestValueHolderPickle(self):
        latest = SecurityLatestValueHolder('x')

        data = dict(aapl=dict(x=1),
                    ibm=dict(x=2))

        latest.push(data)

        f = tempfile.NamedTemporaryFile('w+b', delete=False)
        pickle.dump(latest, f)
        f.close()

        with open(f.name, 'rb') as f2:
            pickled = pickle.load(f2)
            calculated = pickled.value

            for name in data.keys():
                self.assertAlmostEqual(data[name]['x'], calculated[name])

        os.unlink(f.name)

    def testSecurityLtOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') < SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' \\lt ''\\text{y}''")

    def testSecurityLeOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') <= SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' \\le ''\\text{y}''")

    def testSecurityGtOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') > SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' \\gt ''\\text{y}''")

    def testSecurityGeOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') >= SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' \\ge ''\\text{y}''")

    def testSecurityEqOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') == SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' == ''\\text{y}''")

    def testSecurityNeOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') != SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' \\neq ''\\text{y}''")

    def testSecurityAndOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') & SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' \& ''\\text{y}''")

    def testSecurityOrOperatorValueHolderStr(self):
        holder = SecurityLatestValueHolder('x') | SecurityLatestValueHolder('y')
        self.assertEqual(str(holder), "''\\text{x}'' | ''\\text{y}''")



# -*- coding: utf-8 -*-
u"""
Created on 2017-1-13

@author: cheng.li
"""

import unittest
import warnings
from PyFin.Utilities import require
from PyFin.Utilities import warning
from PyFin.Utilities import isClose


class TestAsserts(unittest.TestCase):

    def testPyFinAssert(self):

        with self.assertRaises(ValueError):
            require(1 == 2, ValueError)

    def testPyFinWarning(self):
        with warnings.catch_warnings(record=True) as warning_list:
            warnings.simplefilter("always")
            warning(1 == 2, DeprecationWarning)
            self.assertTrue(any(item.category == DeprecationWarning for item in warning_list))

    def testIsClose(self):
        self.assertTrue(isClose(1.0, 1.000001, 1e-5))
        self.assertTrue(isClose(1e9, 1. + 1e9, 1e-5, 1e-6))
        self.assertFalse(isClose(1., 1. + 1e-5, 1e-6, 1e-6))


# -*- coding: utf-8 -*-
u"""
Created on 2017-6-12

@author: cheng.li
"""

import unittest
from PyFin.DateUtilities.Date import Date
from PyFin.DateUtilities.Period import Period
from PyFin.DateUtilities.Calendar import Calendar
from PyFin.DateUtilities.Schedule import Schedule
from PyFin.DateUtilities.DayCounter import DayCounter


class SingleCase(object):
    def __init__(self, start, end, result, refStart=None, refEnd=None):
        self.start = start
        self.end = end
        self.refStart = refStart
        self.refEnd = refEnd
        self.result = result


class TestDayCounter(unittest.TestCase):
    def testActualActual(self):
        cases = [
            # first example
            SingleCase(Date(2003, 11, 1), Date(2004, 5, 1), 0.500000000000, Date(2003, 11, 1), Date(2004, 5, 1)),
            # short first calculation period(first period)
            SingleCase(Date(1999, 2, 1), Date(1999, 7, 1), 0.410958904110, Date(1998, 7, 1), Date(1999, 7, 1)),
            # short first calculation period(second period)
            SingleCase(Date(1999, 7, 1), Date(2000, 7, 1), 1.000000000000, Date(1999, 7, 1), Date(2000, 7, 1)),
            # long first calculation period(first period)
            SingleCase(Date(2002, 8, 15), Date(2003, 7, 15), 0.915760869565, Date(2003, 1, 15), Date(2003, 7, 15)),
            # long first calculation period(second period) / *Warning:
            # the ISDA case is in disagreement with mktc1198.pdf *
            SingleCase(Date(2003, 7, 15), Date(2004, 1, 15), 0.500000000000, Date(2003, 7, 15), Date(2004, 1, 15)),
            # short final calculation period(penultimate period)
            SingleCase(Date(1999, 7, 30), Date(2000, 1, 30), 0.500000000000, Date(1999, 7, 30), Date(2000, 1, 30)),
            # short final calculation period(final period)
            SingleCase(Date(2000, 1, 30), Date(2000, 6, 30), 0.417582417582, Date(2000, 1, 30), Date(2000, 7, 30))
        ]

        for case in cases:
            dayCounter = DayCounter('Actual/Actual (ISMA)')
            d1 = case.start
            d2 = case.end
            rd1 = case.refStart
            rd2 = case.refEnd

            calculated = dayCounter.yearFraction(d1, d2, rd1, rd2)
            self.assertAlmostEqual(case.result, calculated)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
u"""
Created on 2015-8-8

@author: cheng.li
"""

from PyFin.tests.Math.Timeseries.testNormalizers import TestNormalizers
from PyFin.tests.Math.Timeseries.testTimeseries import TestTimeseries

__all__ = ['TestNormalizers',
           'TestTimeseries']



# -*- coding: utf-8 -*-
u"""
Created on 2015-8-8

@author: cheng.li
"""

__all__ = ['Timeseries',
           'Normalizer']

from PyFin.Math.Timeseries.Timeseries import Timeseries
from PyFin.Math.Timeseries.Normalizers import Normalizer



# -*- coding: utf-8 -*-
u"""
Created on 2015-7-23

@author: cheng.li
"""

from PyFin.Math.Distributions.NormalDistribution import NormalDistribution
from PyFin.Math.Distributions.NormalDistribution import CumulativeNormalDistribution
from PyFin.Math.Distributions.NormalDistribution import InverseCumulativeNormal

__all__ = ['NormalDistribution',
           'CumulativeNormalDistribution',
           'InverseCumulativeNormal']



# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages
from distutils.cmd import Command
from distutils.extension import Extension
import os
import sys
import io
import subprocess
import platform
import numpy as np
from Cython.Build import cythonize
import Cython.Compiler.Options
Cython.Compiler.Options.annotate = True

if "--line_trace" in sys.argv:
    line_trace = True
    profile = True
    print("Build with line trace enabled ...")
    sys.argv.remove("--line_trace")
else:
    line_trace = False
    profile = False

PACKAGE = "PyFin"
NAME = "Finance-Python"
__version__ = "0.9.5"
DESCRIPTION = "PyFin " + __version__
AUTHOR = "cheng li"
AUTHOR_EMAIL = "wegamekinglc@hotmail.com"
URL = 'https://github.com/ChinaQuants/Finance-Python'


def git_version():
    from subprocess import Popen, PIPE
    gitproc = Popen(['git', 'rev-parse', 'HEAD'], stdout=PIPE)
    (stdout, _) = gitproc.communicate()
    return stdout.strip().decode("utf8")


class test(Command):
    description = "test the distribution prior to install"

    user_options = [
        ('test-dir=', None,
         "directory that contains the test definitions"),
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if sys.platform == 'win32':
            command = "coverage run PyFin/tests/testSuite.py& coverage report& coverage html"
        else:
            command = "coverage run PyFin/tests/testSuite.py; coverage report; coverage html"
        process = subprocess.Popen(command, shell=True)
        process.wait()


class version_build(Command):

    description = "test the distribution prior to install"

    user_options = [
        ('test-dir=', None,
         "directory that contains the test definitions"),
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        git_ver = git_version()[:10]
        configFiles = ['PyFin/__init__.py', 'setup.py']

        for configFile in configFiles:
            file_handle = open(configFile, 'r')
            lines = file_handle.readlines()
            newFiles = []
            for line in lines:
                if line.startswith('__version__'):
                    line = line.split('+')[0].rstrip()
                    line = line + " + \"-" + git_ver + "\"\n"
                newFiles.append(line)
            file_handle.close()
            os.remove(configFile)
            file_handle = open(configFile, 'w')
            file_handle.writelines(newFiles)
            file_handle.close()


requirements = "requirements.txt"


ext_modules = [
    "PyFin/Analysis/SeriesValues.pyx",
    "PyFin/Analysis/transformer.pyx",
    "PyFin/Analysis/SecurityValueHolders.pyx",
    "PyFin/Analysis/CrossSectionValueHolders.pyx",
    "PyFin/Analysis/TechnicalAnalysis/StatefulTechnicalAnalysers.pyx",
    "PyFin/Analysis/TechnicalAnalysis/StatelessTechnicalAnalysers.pyx",
    "PyFin/Math/Accumulators/impl.pyx",
    "PyFin/Math/Accumulators/IAccumulators.pyx",
    "PyFin/Math/Accumulators/StatefulAccumulators.pyx",
    "PyFin/Math/Accumulators/StatelessAccumulators.pyx",
    "PyFin/Math/Distributions/NormalDistribution.pyx",
    "PyFin/Math/Distributions/norm.pyx",
    "PyFin/Math/ErrorFunction.pyx",
    "PyFin/Math/MathConstants.pyx",
    "PyFin/Math/udfs.pyx",
    "PyFin/DateUtilities/Calendar.pyx",
    "PyFin/DateUtilities/Date.pyx",
    "PyFin/DateUtilities/Period.pyx",
    "PyFin/DateUtilities/Schedule.pyx",
    "PyFin/Utilities/Asserts.pyx",
    "PyFin/Utilities/Tools.pyx",
    "PyFin/PricingEngines/BlackFormula.pyx",
    "PyFin/PricingEngines/SabrFormulaImpl.pyx",
    "PyFin/PricingEngines/SVIInterpolationImpl.pyx",
    "PyFin/Enums/TimeUnits.pyx",
    "PyFin/Enums/BizDayConventions.pyx",
    "PyFin/Enums/DateGeneration.pyx",
    "PyFin/Enums/Months.pyx",
    "PyFin/Enums/NormalizingType.pyx",
    "PyFin/Enums/OptionType.pyx",
    "PyFin/Enums/Weekdays.pyx"
]


def generate_extensions(ext_modules, line_trace=False):

    extensions = []

    if line_trace:
        print("define cython trace to True ...")
        define_macros = [('CYTHON_TRACE', 1), ('CYTHON_TRACE_NOGIL', 1)]
    else:
        define_macros = []

    for pyxfile in ext_modules:
        ext = Extension(name='.'.join(pyxfile.split('/'))[:-4],
                        sources=[pyxfile],
                        define_macros=define_macros)
        extensions.append(ext)
    return extensions


if platform.system() != "Windows":
    import multiprocessing
    n_cpu = multiprocessing.cpu_count()
else:
    n_cpu = 0

ext_modules_settings = cythonize(generate_extensions(ext_modules, line_trace), 
                                 compiler_directives={"embedsignature": True,
                                                      "linetrace": line_trace,
                                                      "profile": profile,
                                                      "language_level": 3},
                                 nthreads=n_cpu)


setup(
    name=NAME,
    version=__version__,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    include_package_data=False,
    install_requires=io.open(requirements, encoding='utf8').read(),
    classifiers=[],
    cmdclass={"test": test,
              "version_build": version_build},
    ext_modules=ext_modules_settings,
    include_dirs=[np.get_include()],
)
